// HebrewCompare_res.rc
// This file is public domain software.
// Copyright (C) 2018 Katayama Hirofumi MZ <katayama.hirofumi.mz@gmail.com>
//////////////////////////////////////////////////////////////////////////////

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H
#include <windows.h>
#include <windowsx.h>
#include <cmath>

FT_Library freetype;

FT_Face win_face;
FT_Face ros_face;
FT_Face wine_face;

FT_Face win_bold_face;
FT_Face ros_bold_face;
FT_Face wine_bold_face;

WCHAR szText[128];
INT ich = 0, cchText = 0;

//////////////////////////////////////////////////////////////////////////////

typedef struct BITMAPINFODX
{
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[256];
} BITMAPINFODX;

BOOL SaveBitmapToFile(LPCTSTR pszFileName, HBITMAP hbm)
{
    BOOL f;
    BITMAPFILEHEADER bf;
    BITMAPINFODX bi;
    BITMAPINFOHEADER *pbmih;
    DWORD cb;
    DWORD cColors, cbColors;
    HDC hDC;
    HANDLE hFile;
    LPVOID pBits;
    BITMAP bm;

    if (!GetObject(hbm, sizeof(BITMAP), &bm))
        return FALSE;

    pbmih = &bi.bmiHeader;
    ZeroMemory(pbmih, sizeof(BITMAPINFOHEADER));
    pbmih->biSize             = sizeof(BITMAPINFOHEADER);
    pbmih->biWidth            = bm.bmWidth;
    pbmih->biHeight           = bm.bmHeight;
    pbmih->biPlanes           = 1;
    pbmih->biBitCount         = bm.bmBitsPixel;
    pbmih->biCompression      = BI_RGB;
    pbmih->biSizeImage        = bm.bmWidthBytes * bm.bmHeight;

    if (bm.bmBitsPixel < 16)
        cColors = 1 << bm.bmBitsPixel;
    else
        cColors = 0;
    cbColors = cColors * sizeof(RGBQUAD);

    bf.bfType = 0x4d42;
    bf.bfReserved1 = 0;
    bf.bfReserved2 = 0;
    cb = sizeof(BITMAPFILEHEADER) + pbmih->biSize + cbColors;
    bf.bfOffBits = cb;
    bf.bfSize = cb + pbmih->biSizeImage;

    pBits = HeapAlloc(GetProcessHeap(), 0, pbmih->biSizeImage);
    if (pBits == NULL)
        return FALSE;

    f = FALSE;
    hDC = CreateCompatibleDC(NULL);
    if (hDC != NULL)
    {
        if (GetDIBits(hDC, hbm, 0, bm.bmHeight, pBits, (BITMAPINFO*)&bi,
            DIB_RGB_COLORS))
        {
            hFile = CreateFile(pszFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL |
                               FILE_FLAG_WRITE_THROUGH, NULL);
            if (hFile != INVALID_HANDLE_VALUE)
            {
                f = WriteFile(hFile, &bf, sizeof(BITMAPFILEHEADER), &cb, NULL) &&
                    WriteFile(hFile, &bi, sizeof(BITMAPINFOHEADER), &cb, NULL) &&
                    WriteFile(hFile, bi.bmiColors, cbColors, &cb, NULL) &&
                    WriteFile(hFile, pBits, pbmih->biSizeImage, &cb, NULL);
                CloseHandle(hFile);

                if (!f)
                    DeleteFile(pszFileName);
            }
        }
        DeleteDC(hDC);
    }

    HeapFree(GetProcessHeap(), 0, pBits);
    return f;
}

HBITMAP ConvertToGrayscaleBitmap(FT_Bitmap bitmap)
{
    HBITMAP hbm;
    BITMAPINFODX bi;

    ZeroMemory(&bi, sizeof(bi));
    bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = bitmap.width;
    bi.bmiHeader.biHeight = -bitmap.rows;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount = 8;
    bi.bmiHeader.biCompression = BI_RGB;

    for (INT i = 0; i < 256; ++i)
    {
        bi.bmiColors[i].rgbBlue = (BYTE)i;
        bi.bmiColors[i].rgbGreen = (BYTE)i;
        bi.bmiColors[i].rgbRed = (BYTE)i;
        bi.bmiColors[i].rgbReserved = 0;
    }

    LPVOID pvBits;
    hbm = CreateDIBSection(NULL, (BITMAPINFO *)&bi, DIB_RGB_COLORS, &pvBits, NULL, 0);
    if (hbm == NULL)
        return NULL;

    BITMAP bm;
    GetObject(hbm, sizeof(bm), &bm);

    LPBYTE pbBits = (LPBYTE)pvBits;
    for (INT y = 0; y < bitmap.rows; ++y)
    {
        CopyMemory(&pbBits[y * bm.bmWidthBytes],
                   &bitmap.buffer[y * bitmap.pitch],
                   bitmap.width);
        for (INT x = 0; x < bm.bmWidthBytes; ++x)
        {
            pbBits[y * bm.bmWidthBytes + x] = 255 - pbBits[y * bm.bmWidthBytes + x];
        }
    }

    return hbm;
}

//////////////////////////////////////////////////////////////////////////////

FT_Error LoadFace(FT_Face *face, const char *pathname)
{
    return FT_New_Face(freetype, pathname, 0, face);
}

FT_GlyphSlot CreateGlyphSlot(FT_Face face, WCHAR charcode)
{
    FT_Error error;

    FT_UInt glyph_index = FT_Get_Char_Index(face, charcode);
    if (glyph_index == 0)
        glyph_index = FT_Get_Char_Index(face, charcode + 0xF000);   // FT_ENCODING_MS_SYMBOL

    FT_Int32 load_flags = FT_LOAD_DEFAULT;
    error = FT_Load_Glyph(face, glyph_index, load_flags);
    if (error)
    {
        printf("FT_Load_Glyph\n");
        return NULL;
    }

    FT_Render_Mode render_mode = FT_RENDER_MODE_NORMAL;
    error = FT_Render_Glyph(face->glyph, render_mode);
    if (error)
    {
        printf("FT_Render_Glyph\n");
        return NULL;
    }

    return face->glyph;
}

//////////////////////////////////////////////////////////////////////////////

void DoPaint(HWND hwnd, HDC hDC)
{
    RECT rc;
    GetClientRect(hwnd, &rc);
    rc.top += 40;

    SIZE siz;
    siz.cx = rc.right - rc.left;
    siz.cy = rc.bottom - rc.top;

    rc.left += siz.cx / 5;
    rc.bottom -= siz.cy / 5;

    siz.cx = rc.right - rc.left;
    siz.cy = rc.bottom - rc.top;

    INT cxy = GetDlgItemInt(hwnd, edt1, NULL, TRUE);

    POINT pt;
    pt.x = rc.left;
    pt.y = rc.top + cxy;

    if (HPEN hPen = CreatePen(PS_SOLID, 1, RGB(0, 0, 255)))
    {
        SelectObject(hDC, GetStockObject(NULL_BRUSH));

        if (HDC hdcMem = CreateCompatibleDC(hDC))
        {
            INT i = SendDlgItemMessage(hwnd, cmb1, CB_GETCURSEL, 0, 0);
            BOOL bBold = IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED;

            FT_Face face;
            FT_GlyphSlot slot;
            switch (i)
            {
            case 0:
                if (bBold)
                    face = win_bold_face;
                else
                    face = win_face;
                break;
            case 1:
                if (bBold)
                    face = ros_bold_face;
                else
                    face = ros_face;
                break;
            case 2:
                if (bBold)
                    face = wine_bold_face;
                else
                    face = wine_face;
                break;
            }

            INT points = cxy * 72 / GetDeviceCaps(hdcMem, LOGPIXELSY);
            FT_Set_Pixel_Sizes(face, 0, cxy);

            for (INT i = 0; i < face->num_charmaps; ++i)
            {
                FT_CharMap cmap = face->charmaps[i];
                if (cmap->encoding == FT_ENCODING_UNICODE ||
                    cmap->encoding == FT_ENCODING_MS_SYMBOL)
                {
                    FT_Set_Charmap(face, cmap);
                    break;
                }
            }

            slot = CreateGlyphSlot(face, szText[ich]);
            INT cx = (slot->advance.x >> 6);
            FT_Bitmap bitmap = slot->bitmap;

            RECT rc;
            SetRect(&rc, pt.x, pt.y - cxy, pt.x + cx, pt.y);
            FillRect(hDC, &rc, GetStockBrush(WHITE_BRUSH));

            if (HBITMAP hbm = ConvertToGrayscaleBitmap(bitmap))
            {
                HGDIOBJ hbmOld = SelectObject(hdcMem, hbm);
                {
                    BitBlt(hDC, pt.x + slot->bitmap_left, pt.y - slot->bitmap_top,
                           bitmap.width, bitmap.rows, hdcMem, 0, 0, SRCCOPY);
                }
                SelectObject(hdcMem, hbmOld);
                DeleteObject(hbm);
            }

            HGDIOBJ hPenOld = SelectObject(hDC, hPen);
            Rectangle(hDC, pt.x, pt.y - cxy, pt.x + cx + 1, pt.y);
            SelectObject(hDC, hPenOld);

            DeleteDC(hdcMem);
        }

        DeleteObject(hPen);
    }
}

void OnPaint(HWND hwnd)
{
    RECT rc;
    ::GetClientRect(hwnd, &rc);

    PAINTSTRUCT ps;
    HDC hDC = ::BeginPaint(hwnd, &ps);
    if (hDC)
    {
        DoPaint(hwnd, hDC);

        ::EndPaint(hwnd, &ps);
    }
}

void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    RECT rc;
    GetClientRect(hwnd, &rc);
    rc.top += 40;

    SIZE siz;
    siz.cx = rc.right - rc.left;
    siz.cy = rc.bottom - rc.top;

    rc.left += siz.cx / 5;
    rc.bottom -= siz.cy / 5;

    siz.cx = rc.right - rc.left;
    siz.cy = rc.bottom - rc.top;

    INT cxy;
    if (siz.cx < siz.cy)
    {
        cxy = siz.cx;
    }
    else
    {
        cxy = siz.cy;
    }

    SetDlgItemInt(hwnd, edt1, cxy, TRUE);
    ::InvalidateRect(hwnd, NULL, TRUE);
}

void OnPsh1(HWND hwnd)
{
    if (ich > 0)
        --ich;
    WCHAR sz[2];
    sz[0] = szText[ich];
    sz[1] = 0;
    SetDlgItemTextW(hwnd, edt2, sz);
    ::InvalidateRect(hwnd, NULL, TRUE);
}

void OnPsh2(HWND hwnd)
{
    if (ich + 1 < cchText)
        ++ich;
    WCHAR sz[2];
    sz[0] = szText[ich];
    sz[1] = 0;
    SetDlgItemTextW(hwnd, edt2, sz);
    ::InvalidateRect(hwnd, NULL, TRUE);
}

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
    switch (id)
    {
    case IDOK:
    case IDCANCEL:
        EndDialog(hwnd, id);
        break;
    case psh1:
        if (codeNotify == BN_CLICKED)
        {
            OnPsh1(hwnd);
        }
        break;
    case psh2:
        if (codeNotify == BN_CLICKED)
        {
            OnPsh2(hwnd);
        }
        break;
    case cmb1:
        if (codeNotify == CBN_SELCHANGE)
            ::InvalidateRect(hwnd, NULL, TRUE);
        break;
    case edt1:
        if (codeNotify == EN_CHANGE)
        {
            ::InvalidateRect(hwnd, NULL, TRUE);
        }
        break;
    case chx1:
        if (codeNotify == BN_CLICKED)
            ::InvalidateRect(hwnd, NULL, TRUE);
        break;
    }
}

BOOL OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
    SendDlgItemMessage(hwnd, cmb1, CB_ADDSTRING, 0, (LPARAM)TEXT("Win"));
    SendDlgItemMessage(hwnd, cmb1, CB_ADDSTRING, 0, (LPARAM)TEXT("Ros"));
    SendDlgItemMessage(hwnd, cmb1, CB_ADDSTRING, 0, (LPARAM)TEXT("Wine"));
    SendDlgItemMessage(hwnd, cmb1, CB_SETCURSEL, 0, 0);

    SetDlgItemInt(hwnd, edt1, 20, TRUE);

    WCHAR sz[2];
    sz[0] = szText[ich];
    sz[1] = 0;
    SetDlgItemTextW(hwnd, edt2, sz);

    return TRUE;
}

INT_PTR CALLBACK
DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialog);
        HANDLE_MSG(hwnd, WM_SIZE, OnSize);
        HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
        HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
    }
    return 0;
}

const char *GetFontPath(const char *name)
{
    static char path[MAX_PATH];
    GetWindowsDirectoryA(path, MAX_PATH);
    lstrcatA(path, "\\Fonts\\");
    lstrcatA(path, name);
    return path;
}

//////////////////////////////////////////////////////////////////////////////
// Win32 App main function

extern "C" INT WINAPI
WinMain(HINSTANCE   hInstance,
        HINSTANCE   hPrevInstance,
        LPSTR       lpCmdLine,
        INT         nCmdShow)
{
    FT_Error error;

    LoadStringW(hInstance, 1, szText, ARRAYSIZE(szText));
    cchText = lstrlenW(szText);

    error = FT_Init_FreeType(&freetype);
    if (error)
    {
        printf("FT_Init_FreeType\n");
        return 1;
    }

    error = LoadFace(&win_face, ".\\win-tahoma.ttf");
    if (error)
    {
        error = LoadFace(&win_face, GetFontPath("tahoma.ttf"));
        if (error)
        {
            OutputDebugStringA("FT_New_Face (Win)\n");
            return 2;
        }
    }
    error = LoadFace(&win_bold_face, ".\\win-tahomabd.ttf");
    if (error)
    {
        error = LoadFace(&win_bold_face, GetFontPath("tahomabd.ttf"));
        if (error)
        {
            OutputDebugStringA("FT_New_Face (Win, Bold)\n");
            return 5;
        }
    }
    error = LoadFace(&ros_face, ".\\ros-tahoma.ttf");
    if (error)
    {
        OutputDebugStringA("FT_New_Face (ROS)\n");
        return 3;
    }
    error = LoadFace(&ros_bold_face, ".\\ros-tahomabd.ttf");
    if (error)
    {
        OutputDebugStringA("FT_New_Face (ROS, Bold)\n");
        return 6;
    }

    error = LoadFace(&wine_face, ".\\wine-tahoma.ttf");
    if (error)
    {
        OutputDebugStringA("FT_New_Face (Wine)\n");
        return 4;
    }
    error = LoadFace(&wine_bold_face, ".\\wine-tahomabd.ttf");
    if (error)
    {
        OutputDebugStringA("FT_New_Face (Wine, Bold)\n");
        return 7;
    }

    DialogBox(hInstance, MAKEINTRESOURCE(1), NULL, DialogProc);

    FT_Done_Face(win_face);
    FT_Done_Face(ros_face);
    FT_Done_Face(wine_face);
    FT_Done_FreeType(freetype);

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
